#!/usr/bin/env python
# -*- coding: utf-8 -*-

# GIMP plugin to create a drawing look using G'mic Engrave
#      Inspired by a Nidhogg's tutorial posted on Gimpchat
#      modified to create a Van Gogh style

# (c) Dinasset 2020
#
#   History:
#
#   v1.0: 2020-11-20: Initial version
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


import sys, os
import time
import datetime
from gimpfu import *

# define here some constant value, when required
wrkwh=2400
# presets
poster_perc=40
poster_levels=8

############################################################################
# Service routines

#---------------------------------------------------------------------------
# for messages: include, then use: erro_msg or info_msg
#---------------------------------------------------------------------------
def erro_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(0) # box
    gimp.message ("ERROR: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###
def info_msg (msg):
    ###
    currenth = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2) # error console
    gimp.message ("INFO: "+msg)
    pdb.gimp_message_set_handler(currenth) # current
    ###

#---------------------------------------------------------------------------
# from Original drawable create the Output image
#---------------------------------------------------------------------------
def createOutimage(inImage, inLayer,
                        widthheight):
     
    owidth = inImage.width
    oheight = inImage.height
    otype = inImage.base_type
    oname = inLayer.name
    
   # define/calculate the scale factor for the (new) output image -------------------------
    realWidthHeight = owidth+oheight      # real size of the input image
    if widthheight == 0 :   # means the user wants to have output size same as input size
        widthheight = realWidthHeight
    optiWidthHeight = widthheight       # size to work on to get best results
    scale = (optiWidthHeight*1.0)/(realWidthHeight*1.0)      # ratio for scaling

    rwidth=int(owidth*scale)
    rheight=int(oheight*scale)
    
   # create a new image of the decided size ------------------------------------------------
   # --- copy the source layer
   # --- scale
    newImage = pdb.gimp_image_new (rwidth, rheight, otype)                    
    newBG = pdb.gimp_layer_new_from_drawable(inLayer, newImage)     # background layer
    newBG.mode = LAYER_MODE_NORMAL
    newBG.name = "BG Layer from "+oname
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    pdb.gimp_layer_scale(newBG, rwidth, rheight, False)
    pdb.gimp_layer_add_alpha (newBG)    # RGBA
    # remove black values 0.0-22.5
    pdb.gimp_drawable_levels (newBG, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.10, 0.09, 1.00, False)  
    
   # keep a copy of the original background below the one which will be used ----------------
   # --- and set its visibility off (it should not enter in the process)
    Original = newBG.copy()
    newImage.add_layer(Original, 1)     # below
    pdb.gimp_item_set_visible (Original, False)
    Original.name = oname+" (original)"
   
    newBG.opacity = 100.0
    
    return(newImage)
#--------------------------------------------------------------------------- 
    

############################################################################
# Main
#---------------------------------------------------------------------------

def DrawStyleK (inImage, inLayer,
                widthheight, vanGogh,
                vGyellow, doTouchUp,
                doFlatten):
    
    
    # here the logic of the filter
    oname = inLayer.name
    if "." in oname:
        split = oname.split(".", 1)
        oname = split[0]
    
    ts = time.time()
    st = datetime.datetime.fromtimestamp(ts).strftime("%Y-%m-%d %H-%M-%S")
    
    if widthheight == 0 :   # means the user wants to have output size same as input size
        widthheight = inImage.width+inImage.height      # real size of the input image
        
    # for various layers    
    if vanGogh==0:  # weak effect
        scale=2
        pattopa=30.0
        varyingopa=55.0
    elif vanGogh==1:  # medium effect
        scale=3
        pattopa=60.0
        varyingopa=77.0
    elif vanGogh==2:  # strong effect
        scale=6
        pattopa=90.0
        varyingopa=99.0
    # for displace    
    if vanGogh==0:  # weak effect
        divisor=150.0 # chg from 200 18.11
    elif vanGogh==1:  # medium effect
        divisor=100.0
    elif vanGogh==2:  # strong effect
        divisor=50.0
    # for whirl    
    if vanGogh==0:  # weak effect
        whirl=45.0
    elif vanGogh==1:  # medium effect
        whirl=90.0
    elif vanGogh==2:  # strong effect
        whirl=135.0
    # for outcome suffix    
    if vanGogh==0:  # weak effect
        VGsuffix=" (using Van Gogh weak effect)"
    elif vanGogh==1:  # medium effect
        VGsuffix=" (using Van Gogh medium effect)"
    elif vanGogh==2:  # strong effect
        VGsuffix=" (using Van Gogh strong effect)"
     
    newImage = createOutimage (inImage, inLayer, wrkwh)
    #gimp.message("returning from "+str(tocall))
    #pdb.gimp_display_new (newImage)
    #return
    Preprocessed = newImage.layers[0]
    PpW = Preprocessed.width
    PpH = Preprocessed.height
    PpWH = Preprocessed.width+Preprocessed.height
    
    # add grain
    GrainTexture = pdb.gimp_layer_new (newImage, PpW, PpH, RGBA_IMAGE, "Texture",
                                       100.0, LAYER_MODE_GRAIN_MERGE)
    newImage.add_layer (GrainTexture, 0)
    pdb.gimp_context_set_pattern ("3D Green")
    pdb.gimp_drawable_edit_fill (GrainTexture, FILL_PATTERN)
    pdb.gimp_drawable_desaturate (GrainTexture, DESATURATE_LUMA)
    # keep as is by copying it for later displace   
    DisplaceMap = GrainTexture.copy()
    newImage.add_layer(DisplaceMap, 99)
    pdb.gimp_context_set_pattern ("Walnut") # added 18.11
    pdb.gimp_drawable_fill (DisplaceMap, FILL_PATTERN)  # added 18.11
    #pdb.plug_in_spread (newImage, GrainTexture, 5.0, 5.0)   # removed 18.11
    Preprocessed = pdb.gimp_image_merge_down (newImage, GrainTexture, CLIP_TO_IMAGE)
    # displace ------- (van Gogh) -------------------
    pdb.plug_in_displace (newImage, Preprocessed, int(PpW/divisor), int(PpH/divisor),
                          True, True, DisplaceMap, DisplaceMap, 2)  # 1=WRAP, 2=SMEAR
    pdb.gimp_image_remove_layer (newImage, DisplaceMap)
    # check mean
    (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (Preprocessed,
                                  HISTOGRAM_VALUE, 0.0, 1.0)
        
    mofified=False
    if meanl < 110:
        #info_msg("mean luminance = "+str(meanl))
        if meanl < 50:  # preliminary curve to avoid long loop
            pdb.gimp_drawable_curves_spline (Preprocessed, HISTOGRAM_VALUE,
                                             6, ( 0, 0, 0.125, 0.250, 1, 1) )
        elif meanl < 80:  # preliminary curve to avoid long loop
            pdb.gimp_drawable_curves_spline (Preprocessed, HISTOGRAM_VALUE,
                                             6, ( 0, 0, 0.25, 0.50, 1, 1) )
        while meanl < 110:
            pdb.gimp_drawable_levels (Preprocessed, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      1.05, 0.01, 0.99, False)
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (Preprocessed,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("increased mean luminance = "+str(meanl))
    if meanl > 144:
        #info_msg("mean luminance = "+str(meanl))
        while meanl > 144:
            pdb.gimp_drawable_levels (Preprocessed, HISTOGRAM_VALUE, 0.0, 1.0, False,
                                      0.95, 0.01, 0.99, False)
            (meanl,std,med,pix,cnt,per) = pdb.gimp_drawable_histogram (Preprocessed,
                                  HISTOGRAM_VALUE, 0.00, 1.0)
        mofified=True
        #info_msg("reduced mean luminance = "+str(meanl))
    # van Gogh colours    
    pdb.gimp_drawable_hue_saturation (Preprocessed, HUE_RANGE_ALL,
                                      0.0,0.0,100.0,0.0)
        
    #pdb.gimp_display_new (newImage)
    #return
        
    # step: G'MIC Simple Local Contrast
    # [G'MIC] Simple Local Contrast: simplelocalcontrast_p 25,2,50,1,1,1,0,0.9,1,0.9,0,0
    pdb.plug_in_gmic_qt(newImage, Preprocessed, 1,2 ,
        "-v - -simplelocalcontrast_p 25,2,50,1,1,1,1,0.9,1,0.9,0,0 ")
    #                                                            0=no parallel processing   
    Contrasted = newImage.layers[0]
    Contrasted.name = Preprocessed.name+" after G'mic Simple Local Contrast"
        
    #pdb.gimp_display_new (newImage)
    #return
        
    # step: G'MIC Engrave
    # [G'MIC] Engrave: fx_engrave 0,200,10,8,50,0,0,0,10,1,0,0,0,1,0,50,50: 
    pdb.plug_in_gmic_qt(newImage, Contrasted, 1,2,   # was 1,2
        "-v - -fx_engrave 0.08,150,10,8,50,0,0,0,10,1,0,12,0,1,0,50,50 ")      
    Engraved1 = newImage.layers[0]
    Engraved1.name = "- Engrave (positive) - base"
    Engraved1.opacity = varyingopa
    # add B/C
    (meanV,B,C,D,E,F) = pdb.gimp_drawable_histogram (Engraved1, HISTOGRAM_VALUE, 0.0, 1.0)
    if meanV<100:
        pdb.gimp_drawable_brightness_contrast (Engraved1, +0.25, +0.50)
    elif meanV>150:
        pdb.gimp_drawable_brightness_contrast (Engraved1, -0.25, +0.50)
    else:    
        pdb.gimp_drawable_brightness_contrast (Engraved1, +0.50, +0.50)
        
    #pdb.gimp_display_new (newImage)
    #return
     
    #[G'MIC] Camouflage: fx_camouflage 2,32,100,35,35,35,75,75,75,105,105,105,140,140,140 (gray)
        
    Q1=pdb.gimp_layer_new (newImage, PpW/2, PpH/2, RGB_IMAGE,
                          "Q1", 100.0, LAYER_MODE_NORMAL)
    newImage.add_layer (Q1, 0)
                        
    pdb.plug_in_gmic_qt(newImage, Q1, 1,0,   
        "-v - -fx_camouflage "+str(scale)+",32,100,35,35,35,70,70,70,105,105,105,140,140,140 ")      
    pdb.plug_in_whirl_pinch (newImage, Q1, whirl, 0.0, 0.99)   # 60.0 ... 1.0)
    pdb.plug_in_whirl_pinch (newImage, Q1, whirl, 0.0, 0.99)   # repeat
    Q2=Q1.copy()
    newImage.add_layer (Q2, 0)
    Q2.name="Q2"
    pdb.gimp_item_transform_translate (Q2, PpW/2, 0)
    Q3=Q1.copy()
    newImage.add_layer (Q3, 0)
    Q3.name="Q3"
    pdb.gimp_item_transform_translate (Q3, 0, PpH/2)
    Q4=Q1.copy()
    Q4.name="Q4"
    newImage.add_layer (Q4, 0)
    pdb.gimp_item_transform_translate (Q4, PpW/2, PpH/2)
    # new on centre
    Q5=Q1.copy()
    Q5.name="Q5"
    newImage.add_layer (Q5, 0)
    pdb.gimp_item_transform_translate (Q5, PpW/4, PpH/4)
    
    pdb.gimp_selection_none (newImage)
    
    Q54 = pdb.gimp_image_merge_down (newImage, Q5, CLIP_TO_IMAGE)
    Q543 = pdb.gimp_image_merge_down (newImage, Q54, CLIP_TO_IMAGE)
    Q5432 = pdb.gimp_image_merge_down (newImage, Q543, CLIP_TO_IMAGE)
    Patt1 = pdb.gimp_image_merge_down (newImage, Q5432, CLIP_TO_IMAGE)
    Patt1.mode = LAYER_MODE_GRAIN_MERGE
    Patt1.name = "- Pattern (positive) - modifier"
    # apply colours of the source
    ForColours1 = Contrasted.copy()
    newImage.add_layer (ForColours1, 0)
    ForColours1.mode = LAYER_MODE_HSL_COLOR
    Patt1 = pdb.gimp_image_merge_down (newImage, ForColours1, CLIP_TO_IMAGE)
    Patt1.mode = LAYER_MODE_GRAIN_MERGE # confirm
    Patt1.opacity = varyingopa
     
    #pdb.gimp_display_new (newImage)
    #return

    PattEngrave1 = pdb.gimp_layer_new_from_visible (newImage, newImage,
                                                    "Engraved and patterned (positive) ")
    newImage.add_layer (PattEngrave1, 0)
    PattEngrave1.mode = LAYER_MODE_NORMAL # temporary
    PattEngrave1.opacity = varyingopa
    
    #pdb.gimp_display_new (newImage)
    #return
    
    # now, repetition of the last steps for inverted colours
    Inverted = Contrasted.copy()
    newImage.add_layer (Inverted, 0)
    pdb.plug_in_vinvert (newImage, Inverted)
    Inverted.name = Contrasted.name+" (inverted)"
    # dark look
    pdb.plug_in_gmic_qt(newImage, Inverted, 1,2,  
        "-v - -fx_engrave 0.08,150,10,8,50,0,0,0,10,1,0,12,0,1,0,50,50 ")    
    Engraved2 = newImage.layers[0]
    Engraved2.name = "- Engrave (negative) - base"
    Engraved2.mode = LAYER_MODE_SOFTLIGHT
    Engraved2.opacity = varyingopa
     
    #[G'MIC] Camouflage: fx_camouflage 2,32,0,85,170,255,85,170,127,255,85,127,255,170,0
    #[G'MIC] Camouflage: fx_camouflage 2,32,100,35,35,35,75,75,75,105,105,105,140,140,140 (gray)
    scale=2 # force
    Q1I=pdb.gimp_layer_new (newImage, PpW/2, PpH/2, RGB_IMAGE,  # chg -10
                          "Q1I", 100.0, LAYER_MODE_NORMAL)
    newImage.add_layer (Q1I, 0)
                        
    pdb.plug_in_gmic_qt(newImage, Q1I, 1,0,   
        "-v - -fx_camouflage "+str(scale)+",32,100,35,35,35,70,70,70,105,105,105,140,140,140 ")      
    pdb.plug_in_whirl_pinch (newImage, Q1I, -whirl, 0.0, 0.99)   # 60.0 ... 1.0)chg from 0.95
    pdb.plug_in_whirl_pinch (newImage, Q1I, -whirl, 0.0, 0.99)   # repeat, chg from 0.95
    Q2I=Q1I.copy()
    newImage.add_layer (Q2I, 0)
    Q2I.name="Q2I"
    pdb.gimp_item_transform_translate (Q2I, PpW/2+5, 5) # chg +5
    Q3I=Q1I.copy()
    newImage.add_layer (Q3I, 0)
    Q3I.name="Q3I"
    pdb.gimp_item_transform_translate (Q3I, 5, PpH/2+5)
    Q4I=Q1I.copy()
    Q4I.name="Q4I"
    newImage.add_layer (Q4I, 0)
    pdb.gimp_item_transform_translate (Q4I, PpW/2+5, PpH/2+5)
    # new on centre
    Q5I=Q1I.copy()
    Q5I.name="Q5I"
    newImage.add_layer (Q5I, 0)
    pdb.gimp_item_transform_translate (Q5I, PpW/4+5, PpH/4+5)
    
    pdb.gimp_selection_none (newImage)
    
    Q54I = pdb.gimp_image_merge_down (newImage, Q5I, CLIP_TO_IMAGE)
    Q543I = pdb.gimp_image_merge_down (newImage, Q54I, CLIP_TO_IMAGE)
    Q5432I = pdb.gimp_image_merge_down (newImage, Q543I, CLIP_TO_IMAGE)
    Patt2 = pdb.gimp_image_merge_down (newImage, Q5432I, CLIP_TO_IMAGE)
    Patt2.mode = LAYER_MODE_BURN    # chg from LINEAR_BURN
    Patt2.name = "- Pattern (negative) - modifier"
    Patt2.opacity = varyingopa/2   # for this layer just the half
    pdb.plug_in_vinvert (newImage, Patt2)
    # apply colours of the source
    ForColours2 = Contrasted.copy()
    newImage.add_layer (ForColours2, 0)
    ForColours2.mode = LAYER_MODE_HSL_COLOR
     
    #pdb.gimp_display_new (newImage)
    #return
    
    Posterized = Preprocessed.copy()
    pos = pdb.gimp_image_get_item_position (newImage, Contrasted)
    newImage.add_layer (Posterized, pos)
    pdb.gimp_drawable_posterize (Posterized, poster_levels)          
    Posterized.opacity = poster_perc*1.0                      
    Posterized.mode = LAYER_MODE_NORMAL
    Posterized.name = "Preprocessed copy, posterized"
    
    pdb.gimp_image_remove_layer (newImage, Contrasted)
    pdb.gimp_image_remove_layer (newImage, Inverted)
    
    PattEngrave1.mode=LAYER_MODE_MULTIPLY   # final
    PattEngrave1.opacity=95.0
        
    if vGyellow==True:
        Yellow = pdb.gimp_layer_new (newImage, PpW, PpH, RGBA_IMAGE,
                                     "Van Gogh yellow tone", 100.0,
                                     LAYER_MODE_NORMAL)
        newImage.add_layer (Yellow, 0)
        pdb.gimp_context_set_foreground ((219,202,26,255))  # yellow chrome
        pdb.gimp_drawable_fill (Yellow, FILL_FOREGROUND)
        Yellow.opacity = 50.0
        # addition 20.11
        mask = pdb.gimp_layer_create_mask (Preprocessed, ADD_MASK_COPY)
        pdb.gimp_layer_add_mask (Yellow, mask)
        Yellow.mode = LAYER_MODE_LINEAR_LIGHT

        
    # final touch up 
    postfix = ""
    if doTouchUp==True:
        postfix=", with mask on top"
        TopOver = Preprocessed.copy()
        newImage.add_layer (TopOver, 0)
        TopOver.mode = LAYER_MODE_NORMAL
        TopOver.opacity = 100.0
        mask = pdb.gimp_layer_create_mask (TopOver, ADD_MASK_BLACK)
        pdb.gimp_layer_add_mask (TopOver, mask)
        # set foreground to a light gray for user brush strokes
        pdb.gimp_context_set_foreground ((200,200,200,255))
    
    # final step
    if doFlatten==True:
        if doTouchUp==True:
            pdb.gimp_item_set_visible (TopOver, False)  # not to be merged
        outcome = pdb.gimp_image_merge_visible_layers (newImage, CLIP_TO_BOTTOM_LAYER)
        outcome.name = st+" "+oname+" as a fake VanGogh engraved drawing, option "+postfix
        if doTouchUp==True:
            pdb.gimp_item_set_visible (TopOver, True)
    else:    
        outcome = pdb.gimp_layer_new_from_visible (newImage, newImage, "Outcome")
        newImage.add_layer (outcome, 0)
        pdb.gimp_item_set_visible (outcome, False)
    outcome.name = st+oname+" as a fake VanGogh engraved drawing"+VGsuffix+postfix
    uratio = widthheight*1.0/PpWH*1.0
    uwidth = int(PpW*uratio); uheight = int(PpH*uratio)
    pdb.gimp_image_scale (newImage, uwidth, uheight)   # to user desired size
    pdb.gimp_display_new (newImage)

### Registration
whoiam="\n"+os.path.abspath(sys.argv[0])
author="Nidhogg/dinasset"
menu="<Image>/Diego/Drawing/"
copyrightYear="2020"
desc="Draw Effect Style K ..."

register(
    "DrawStyleK",
    desc+whoiam,desc,
    author,author,
    copyrightYear,
    desc+"...",
    "*",
    [
        (PF_IMAGE, "image", "Input image", None),
        (PF_DRAWABLE, "drawable", "Input drawable", None),
        (PF_ADJUSTMENT, "newsize", "Width+Height of New Image"+
                        "\n (0=same as input)", 3000, (0, 7200, 100)),
        (PF_OPTION, "vangogh", "Van Gogh effect", 1, ("weak",     
                                                      "medium",
                                                      "strong",
                                                     )),
        (PF_TOGGLE, "vgyellow", "Apply Van Gogh yellow?", True),
        (PF_TOGGLE, "touchup", "Touchup final outcome?", True),
        (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    DrawStyleK,
    menu=menu
)

main()
